@@ -15,14 +15,13 @@ class FilePersistence(Persister):
 
     def read(self, task_id):
         acc = v()
-        with open(self.output_file, 'r') as f:
+        with open(self.output_file, "r") as f:
             for line in f:
                 parsed = json.loads(line, object_hook=json_deserializer)
-                if parsed['task_id'] == task_id:
+                if parsed["task_id"] == task_id:
                     acc = acc.append(Event.create(parsed))
         return acc
 
     def write(self, event):
-        with open(self.output_file, 'a+') as f:
-            f.write("{}\n".format(json.dumps(
-                thaw(event), default=json_serializer)))
+        with open(self.output_file, "a+") as f:
+            f.write("{}\n".format(json.dumps(thaw(event), default=json_serializer)))
